# encoding: utf-8
# This file is distributed under Mmtrix's license terms.


DependencyDetection.defer do
  named :data_mapper

  depends_on do
    defined?(::DataMapper) &&
      defined?(::DataMapper::Model) &&
      defined?(::DataMapper::Resource) &&
      defined?(::DataMapper::Collection)
  end

  executes do
    ::Mmtrix::Agent.logger.info 'Installing DataMapper instrumentation'
    require 'mmtrix/agent/datastores/metric_helper'
  end

  executes do
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Model, :get
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Model, :first
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Model, :last
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Model, :all

    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Model, :create
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Model, :create!
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Model, :update
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Model, :update!
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Model, :destroy
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Model, :destroy!

    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Model, :aggregate
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Model, :find
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Model, :find_by_sql
  end

  executes do
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Resource, :update
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Resource, :update!
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Resource, :save
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Resource, :save!
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Resource, :destroy
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Resource, :destroy!
  end

  executes do
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Collection, :get
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Collection, :first
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Collection, :last
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Collection, :all

    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Collection, :lazy_load

    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Collection, :create
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Collection, :create!
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Collection, :update
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Collection, :update!
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Collection, :destroy
    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Collection, :destroy!

    Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Collection, :aggregate
  end

  executes do
    # Catch direct SQL calls that bypass CRUD
    if defined?(::DataMapper::Adapters::DataObjectsAdapter)
      Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Adapters::DataObjectsAdapter, :select, true
      Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Adapters::DataObjectsAdapter, :execute, true
    end
  end

  executes do
    # DM::Validations overrides Model#create, so we patch it here as well
    if defined?(::DataMapper::Validations::ClassMethods)
      Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Validations::ClassMethods, :create
    end
  end

  executes do
    # DM::Transaction calls commit() twice, so potentially shows up twice.
    if defined?(::DataMapper::Transaction)
      Mmtrix::Agent::DataMapperTracing.add_tracer ::DataMapper::Transaction, :commit, true
    end
  end

  executes do
    if defined?(::DataObjects::Connection)
      ::DataObjects::Connection.class_eval do
        include ::Mmtrix::Agent::Instrumentation::DataMapperInstrumentation
      end
    end
  end
end

module Mmtrix
  module Agent
    module DataMapperTracing
      def self.add_tracer(clazz, method_name, operation_only = false)
        clazz.class_eval do
          if method_defined?(method_name) || private_method_defined?(method_name)
            define_method("#{method_name}_with_mmtrix",
                          Mmtrix::Agent::DataMapperTracing.method_body(clazz, method_name, operation_only))

            alias_method "#{method_name}_without_mmtrix", method_name
            alias_method method_name, "#{method_name}_with_mmtrix"
          end
        end
      end

      DATA_MAPPER = "DataMapper".freeze

      def self.method_body(clazz, method_name, operation_only)
        use_model_name   = Mmtrix::Helper.instance_methods_include?(clazz, :model)
        metric_operation = method_name.to_s.gsub(/[!?]/, "")

        Proc.new do |*args, &blk|
          begin
            if operation_only
              # Used by direct SQL, like ::DataMapper::Adapters::DataObjectsAdapter#select
              name = nil
            elsif use_model_name
              # Used by ::DataMapper::Collection to get contained model name
              name = self.model.name
            elsif self.is_a?(Class)
              # Used by class-style access, like Model.first()
              name = self.name
            else
              # Used by instance-style access, like model.update(attr: "new")
              name = self.class.name
            end

            metrics = Mmtrix::Agent::Datastores::MetricHelper.metrics_for(
              DATA_MAPPER,
              metric_operation,
              name)

            Mmtrix::Agent::MethodTracer.trace_execution_scoped(metrics) do
              self.send("#{method_name}_without_mmtrix", *args, &blk)
            end
          end
        end
      end
    end

    module Instrumentation
      module DataMapperInstrumentation
        # Unlike in AR, log is called in DM after the query actually ran,
        # complete with metrics.  Since DO has already calculated the
        # duration, there's nothing more to measure, so just record and log.
        #
        # We rely on the assumption that all possible entry points have been
        # hooked with tracers, ensuring that notice_sql attaches this SQL to
        # the proper call scope.
        def log(msg) #THREAD_LOCAL_ACCESS
          state = Mmtrix::Agent::TransactionState.tl_get
          return unless state.is_execution_traced?

          duration = msg.duration / 1000000.0
          Mmtrix::Agent.instance.transaction_sampler.notice_sql(msg.query, nil, duration, state)
        ensure
          super
        end
      end
    end
  end
end
