# encoding: utf-8
# This file is distributed under Mmtrix's license terms.


require 'mmtrix/agent/parameter_filtering'

module Mmtrix
  module Agent
    module Instrumentation
      module GrapeInstrumentation
        extend self

        API_ENDPOINT   = 'api.endpoint'.freeze
        FORMAT_REGEX   = /\(\/?\.:format\)/.freeze
        VERSION_REGEX  = /:version(\/|$)/.freeze
        EMPTY_STRING   = ''.freeze
        MIN_VERSION    = VersionNumber.new("0.2.0")

        def handle_transaction(endpoint, class_name)
          return unless endpoint && route = endpoint.route
          name_transaction(route, class_name)
          capture_params(endpoint)
        end

        def name_transaction(route, class_name)
          txn_name = name_for_transaction(route, class_name)
          node_name = "Middleware/Grape/#{class_name}/call"
          Transaction.set_default_transaction_name(txn_name, :grape, node_name)
        end

        def name_for_transaction(route, class_name)
          action_name = route.route_path.sub(FORMAT_REGEX, EMPTY_STRING)
          method_name = route.route_method

          if route.route_version
            action_name = action_name.sub(VERSION_REGEX, EMPTY_STRING)
            "#{class_name}-#{route.route_version}#{action_name} (#{method_name})"
          else
            "#{class_name}#{action_name} (#{method_name})"
          end
        end

        def capture_params(endpoint)
          txn = Transaction.tl_current
          env = endpoint.request.env
          params = ParameterFiltering::apply_filters(env, endpoint.params)
          params.delete("route_info")
          txn.filtered_params = params
          txn.merge_request_parameters(params)
        end
      end
    end
  end
end

DependencyDetection.defer do
  # Why not just :grape? mmtrix-grape used that name already, and while we're
  # not shipping yet, overloading the name interferes with the plugin.
  named :grape_instrumentation

  depends_on do
    ::Mmtrix::Agent.config[:disable_grape] == false
  end

  depends_on do
    defined?(::Grape::VERSION) &&
      ::Mmtrix::VersionNumber.new(::Grape::VERSION) >= ::Mmtrix::Agent::Instrumentation::GrapeInstrumentation::MIN_VERSION
  end

  depends_on do
    begin
      if defined?(Bundler) && Bundler.rubygems.all_specs.map(&:name).include?("mmtrix-grape")
        ::Mmtrix::Agent.logger.info("Not installing Mmtrix supported Grape instrumentation because the third party mmtrix-grape gem is present")
        false
      else
        true
      end
    rescue => e
      ::Mmtrix::Agent.logger.info("Could not determine if third party mmtrix-grape gem is installed", e)
      true
    end
  end

  executes do
    Mmtrix::Agent.logger.info 'Installing Mmtrix supported Grape instrumentation'
    instrument_call
  end

  def instrument_call
    ::Grape::API.class_eval do
      def call_with_mmtrix(env)
        begin
          response = call_without_mmtrix(env)
        ensure
          begin
            endpoint = env[::Mmtrix::Agent::Instrumentation::GrapeInstrumentation::API_ENDPOINT]
            ::Mmtrix::Agent::Instrumentation::GrapeInstrumentation.handle_transaction(endpoint, self.class.name)
          rescue => e
            ::Mmtrix::Agent.logger.warn("Error in Grape instrumentation", e)
          end
        end

        response
      end

      alias_method :call_without_mmtrix, :call
      alias_method :call, :call_with_mmtrix
    end
  end

end
